# -*- coding: utf-8 -*-
"""----------------------------------------------------------------------------
Author:
    Huang Quanyong (wo1fSea)
    quanyongh@foxmail.com
Date:
    2017/4/26
Description:
    AtlasInterface.py
----------------------------------------------------------------------------"""

from ..Utils import ATLAS_FORMAT_PLIST  # , ATLAS_FORMAT_JSON

MAX_RANK = 2 ** 32
MAX_WIDTH = 1024 * 16
MAX_HEIGHT = 1024 * 16


class AtlasInterface(object):
    """
    interface of atlas
    """

    def __init__(self, width=1, height=1, max_width=MAX_WIDTH, max_height=MAX_HEIGHT,
                 force_square=False, border_padding=0, shape_padding=0, inner_padding=0):
        if force_square:
            width = height = max(width, height)
            max_width = max_height = max(max_width, max_height)

        self.size = (width, height)
        self.max_size = (max_width, max_height)

        self.border_padding = border_padding
        self.shape_padding = shape_padding
        self.inner_padding = inner_padding

        self.force_square = force_square

        self.image_rect_list = []

    def dump_plist(self, texture_file_name="", input_base_path=None, atlas_format=ATLAS_FORMAT_PLIST, trim_sprite_names=False):
        import os

        plist_data = {}

        frames = {}
        for image_rect in self.image_rect_list:
            width, height = (image_rect.width - 2 * image_rect.extrude_size, image_rect.height - 2 * image_rect.extrude_size) \
                if not image_rect.rotated else (image_rect.height - 2 * image_rect.extrude_size, image_rect.width - 2 * image_rect.extrude_size)

            center_offset = (0, 0)
            if image_rect.trimmed:
                center_offset = (image_rect.source_box[0] + width / 2. - image_rect.source_size[0] / 2.,
                                 - (image_rect.source_box[1] + height / 2. - image_rect.source_size[1] / 2.))

            path = image_rect.image_path
            if input_base_path is None:
                _, path = os.path.split(path)
            else:
                path = os.path.relpath(os.path.abspath(
                    path), os.path.abspath(input_base_path))
                path = path.replace('\\', '/') # for windows path

            frame_key = path
            if trim_sprite_names:
                frame_key, _ = os.path.splitext(frame_key)
            if atlas_format == ATLAS_FORMAT_PLIST:
                frames[frame_key] = dict(
                    frame="{{%d,%d},{%d,%d}}" % (
                        image_rect.x, image_rect.y, width, height),
                    offset="{%d,%d}" % center_offset,
                    rotated=bool(image_rect.rotated),
                    sourceColorRect="{{%d,%d},{%d,%d}}" % (
                        image_rect.source_box[0], image_rect.source_box[1], width, height),
                    sourceSize="{%d,%d}" % image_rect.source_size,
                )
            else:
                frames[frame_key] = dict(
                    frame=dict(x=image_rect.x, y=image_rect.y,
                               w=width, h=height),
                    rotated=bool(image_rect.rotated),
                    trimed=bool(image_rect.trimmed),
                    spriteSourceSize=dict(
                        x=image_rect.source_box[0], y=image_rect.source_box[1],
                        w=image_rect.source_box[2], h=image_rect.source_box[3]),
                    sourceSize=dict(
                        w=image_rect.source_size[0], h=image_rect.source_size[1])
                )

        plist_data["frames"] = frames
        if atlas_format == ATLAS_FORMAT_PLIST:
            plist_data["metadata"] = dict(
                format=int(2),
                textureFileName=texture_file_name,
                realTextureFileName=texture_file_name,
                size="{%d,%d}" % self.size,
            )
        else:
            plist_data["meta"] = dict(
                image=texture_file_name,
                format="RGBA8888",
                size=dict(w=self.size[0], h=self.size[1]),
                scale=1,
            )

        return plist_data

    def dump_image(self, bg_color=0xffffffff):
        from PIL import Image
        packed_image = Image.new('RGBA', self.size, bg_color)

        for image_rect in self.image_rect_list:
            image = image_rect.image.crop()
            if image_rect.rotated:
                image = image.transpose(Image.ROTATE_270)
            packed_image.paste(
                image, (image_rect.left, image_rect.top, image_rect.right, image_rect.bottom))

        return packed_image



# -*- coding: utf-8 -*-
"""----------------------------------------------------------------------------
Author:
    Huang Quanyong (wo1fSea)
    quanyongh@foxmail.com
Date:
    2016/10/19
Description:
    GuillotinePacker.py
----------------------------------------------------------------------------"""

from ..MaxRectsPacker.MaxRectsPacker import MaxRectsPacker
from ..GuillotinePacker.GuillotineAtlas import GuillotineAtlas


class GuillotinePacker(MaxRectsPacker):
    """
    a bin packer using guillotine algorithm
    """

    ATLAS_TYPE = GuillotineAtlas



# -*- coding: utf-8 -*-
"""----------------------------------------------------------------------------
Author:
    Huang Quanyong (wo1fSea)
    quanyongh@foxmail.com
Date:
    2016/10/19
Description:
    MaxRectsAtlas.py
----------------------------------------------------------------------------"""

from ..Rect import Rect
from ..PackerInterface.AtlasInterface import AtlasInterface

MAX_RANK = 2 ** 32
MAX_WIDTH = 1024 * 16
MAX_HEIGHT = 1024 * 16


class MaxRectsAtlas(AtlasInterface):
    """
    the max rects data structure used in max rects bin pack algorithm
    """

    # EXPAND STRATEGY
    EXPAND_BOTH = 0
    EXPAND_WIDTH = 1
    EXPAND_HEIGHT = 2
    EXPAND_SHORT_SIDE = 3
    EXPAND_LONG_SIDE = 4

    # RANK STRATEGY
    RANK_BSSF = 0
    RANK_BLSF = 1
    RANK_BAF = 3

    def __init__(self, *args, **kwargs):
        super(MaxRectsAtlas, self).__init__(*args, **kwargs)

        width, height = self.size
        self.max_rect_list = [Rect(0 + self.border_padding,
                                   0 + self.border_padding,
                                   width - 2 * self.border_padding,
                                   height - 2 * self.border_padding)]

    def _is_in_max_size(self, size):
        return size[0] <= self.max_size[0] and size[1] <= self.max_size[1]

    def expand(self, method=EXPAND_SHORT_SIDE):
        if self.force_square:
            method = self.EXPAND_BOTH

        if method == MaxRectsAtlas.EXPAND_BOTH:
            new_size = (self.size[0] * 2, self.size[1] * 2)
        elif method == MaxRectsAtlas.EXPAND_WIDTH:
            new_size = (self.size[0] * 2, self.size[1])
        elif method == MaxRectsAtlas.EXPAND_HEIGHT:
            new_size = (self.size[0], self.size[1] * 2)
        elif method == MaxRectsAtlas.EXPAND_SHORT_SIDE:
            if self.size[0] <= self.size[1]:
                new_size = (self.size[0] * 2, self.size[1])
            else:
                new_size = (self.size[0], self.size[1] * 2)
        elif method == MaxRectsAtlas.EXPAND_LONG_SIDE:
            if self.size[0] >= self.size[1]:
                new_size = (self.size[0] * 2, self.size[1])
            else:
                new_size = (self.size[0], self.size[1] * 2)
        else:
            raise ValueError("Unexpected Method")

        if not self._is_in_max_size(new_size):
            return False

        old_size = self.size
        self.size = new_size

        for max_rect in self.max_rect_list:
            if max_rect.right == old_size[0] - self.border_padding:
                max_rect.right = self.size[0] - self.border_padding
            if max_rect.bottom == old_size[1] - self.border_padding:
                max_rect.bottom = self.size[1] - self.border_padding

        if old_size[0] != self.size[0]:
            new_rect = Rect(old_size[0] - self.border_padding,
                            0 + self.border_padding,
                            self.size[0] - old_size[0],
                            self.size[1] - 2 * self.border_padding)
            self.max_rect_list.append(new_rect)

        if old_size[1] != self.size[1]:
            new_rect = Rect(0 + self.border_padding,
                            old_size[1] - self.border_padding,
                            self.size[0] - 2 * self.border_padding,
                            self.size[1] - old_size[1])
            self.max_rect_list.append(new_rect)

        self.max_rect_list = list(
            filter(self._max_rect_list_pruning, self.max_rect_list))

        return True

    def cut(self, main_rect, sub_rect):
        if not main_rect.is_overlaped(sub_rect):
            return [main_rect]

        result = []
        if main_rect.left < sub_rect.left:
            tmp = main_rect.clone()
            tmp.right = sub_rect.left
            if tmp.area > 0:
                result.append(tmp)
        if main_rect.top < sub_rect.top:
            tmp = main_rect.clone()
            tmp.bottom = sub_rect.top
            if tmp.area > 0:
                result.append(tmp)
        if main_rect.right > sub_rect.right:
            tmp = main_rect.clone()
            tmp.left = sub_rect.right
            if tmp.area > 0:
                result.append(tmp)
        if main_rect.bottom > sub_rect.bottom:
            tmp = main_rect.clone()
            tmp.top = sub_rect.bottom
            if tmp.area > 0:
                result.append(tmp)

        return result

    def rank(self, main_rect, sub_rect, method=RANK_BSSF):
        """
        rank
        :param main_rect:
        :param sub_rect:
        :param method:
        :return:
        """
        if method == self.RANK_BSSF:
            tmp = main_rect.width - sub_rect.width if main_rect.width < main_rect.height \
                else main_rect.height - sub_rect.height
        elif method == self.RANK_BLSF:
            tmp = main_rect.width - sub_rect.width if main_rect.width > main_rect.height \
                else main_rect.height - sub_rect.height
        elif method == self.RANK_BAF:
            tmp = main_rect.area - sub_rect.area

        assert tmp < MAX_RANK
        if tmp < 0 \
                or main_rect.width - sub_rect.width < 2 * self.inner_padding + self.shape_padding \
                or main_rect.height - sub_rect.height < 2 * self.inner_padding + self.shape_padding:
            return MAX_RANK
        else:
            return tmp

    def find_best_rank(self, image_rect, enable_rotated=False):
        if enable_rotated and image_rect.width != image_rect.height:
            return self.find_best_rank_with_rotate(image_rect)
        else:
            return self.find_best_rank_without_rotate(image_rect)

    def find_best_rank_without_rotate(self, image_rect):
        best_rank = MAX_RANK
        best_index = -1
        for i, rect in enumerate(self.max_rect_list):
            rank = self.rank(rect, image_rect)
            if rank < best_rank:
                best_rank = rank
                best_index = i
        return best_index, best_rank, False

    def find_best_rank_with_rotate(self, image_rect):
        image_rect_r = image_rect.clone()
        image_rect_r.rotate()

        index, rank, _ = self.find_best_rank_without_rotate(image_rect)
        index_r, rank_r, _ = self.find_best_rank_without_rotate(image_rect_r)

        if rank < rank_r:
            return index, rank, False
        else:
            return index_r, rank, True

    def place_image_rect(self, rect_index, image_rect):
        rect = self.max_rect_list[rect_index]

        image_rect.x, image_rect.y = rect.x + \
            self.inner_padding, rect.y + self.inner_padding

        fake_image_rect = image_rect.clone()
        fake_image_rect.left -= self.inner_padding
        fake_image_rect.right += self.inner_padding + self.shape_padding
        fake_image_rect.top -= self.inner_padding
        fake_image_rect.bottom += self.inner_padding + self.shape_padding

        _max_rect_list = []
        _new_max_rect_list = []
        for i, rect in enumerate(self.max_rect_list):
            if fake_image_rect.is_overlaped(rect):
                _new_max_rect_list.extend(self.cut(rect, fake_image_rect))
            else:
                _max_rect_list.append(rect)

        self.max_rect_list = _new_max_rect_list
        self.max_rect_list = list(
            filter(self._max_rect_list_pruning, _new_max_rect_list))
        self.max_rect_list.extend(_max_rect_list)

        self.image_rect_list.append(image_rect)

    def _max_rect_list_pruning(self, rect):
        for max_rect in self.max_rect_list:
            if rect in max_rect and rect != max_rect:
                return False

        return True



# -*- coding: utf-8 -*-
"""----------------------------------------------------------------------------
Author:
    Huang Quanyong (wo1fSea)
    quanyongh@foxmail.com
Date:
    2016/10/19
Description:
    Utils.py
----------------------------------------------------------------------------"""

import sys
import inspect
if sys.version_info.major > 2:
    xrange = range

SUPPORTED_IMAGE_FORMAT = [".png", ".jpg", ".bmp"]
ATLAS_FORMAT_PLIST = "plist"
ATLAS_FORMAT_JSON = "json"
ATLAS_FORMAT_CSV = "csv"


def load_images_from_paths(image_path_list):
    """
    load image form paths
    :param image_path_list: image paths list
    :return: ImageRect list
    """
    from .ImageRect import ImageRect

    image_rect_list = []
    for file_path in image_path_list:
        image_rect = ImageRect(file_path)
        image_rect_list.append(image_rect)

    return image_rect_list


def load_images_from_dir(dir_path):
    """
    load all images from a directory
    :param dir_path: directory path
    :return: ImageRect list
    """
    import os

    image_rect_path = []
    for root, dirs, files in os.walk(dir_path):
        for f in files:
            file_path = os.path.join(root, f)
            _, ext = os.path.splitext(f)
            if ext.lower() in SUPPORTED_IMAGE_FORMAT:
                image_rect_path.append(file_path)

    return load_images_from_paths(image_rect_path)


def get_atlas_data_ext(atlas_format):
    """
    get default file extension for selected type of atlas data file
    :param atlas_format: atlas data file format
    :return:
    """
    if atlas_format == ATLAS_FORMAT_PLIST:
        return '.plist'
    elif atlas_format == ATLAS_FORMAT_JSON:
        return '.json'
    elif atlas_format == ATLAS_FORMAT_CSV:
        return '.csv'
    elif callable(atlas_format):
        parameters = inspect.signature(atlas_format).parameters
        required_args = sum(1 for param in parameters.values()
                            if param.default is param.empty)
        if len(parameters) >= 2 and required_args <= 2:
            return '.txt'

    raise ValueError("Unsupported file format: %s" % atlas_format)


def save_atlas_data(data_dict, file_path, atlas_format):
    """
    save a atlas data in selected format. plist or json supported
    :param data_dict: dict data
    :param file_path: file path to save
    :param atlas_format: atlas data file format
    :return:
    """
    if atlas_format == ATLAS_FORMAT_PLIST:
        return save_plist(data_dict, file_path)
    elif atlas_format == ATLAS_FORMAT_JSON:
        return save_json(data_dict, file_path)
    elif atlas_format == ATLAS_FORMAT_CSV:
        return save_csv(data_dict, file_path)
    elif callable(atlas_format):
        parameters = inspect.signature(atlas_format).parameters
        required_args = sum(1 for param in parameters.values()
                            if param.default is param.empty)
        if len(parameters) >= 2 and required_args <= 2:
            return atlas_format(data_dict, file_path)

    raise ValueError("Unsupported file format: %s" % atlas_format)


def save_csv(data_dict, file_path):
    """
    save a dict as a csv
    :param data_dict: dict data
    :param file_path: csv file path to save
    :return:
    """
    with open(file_path, 'w') as fp:
        for name, data in data_dict['frames'].items():
            frame = data['frame']
            source = data['spriteSourceSize']

            # fp.write(f'{name},{frame["x"]},{frame["y"]},{frame["w"]},{frame["h"]},'
            #          f'{source["x"]},{source["y"]},{source["w"]},{source["h"]},'
            #          f'{data["rotated"]},{data["trimed"]}\n')
            fp.write('%s,%d,%d,%d,%d,%d,%d,%d,%d,%s,%s\n' % (name, frame["x"], frame["y"], frame["w"], frame["h"],
                                                             source["x"], source["y"], source["w"], source["h"],
                                                             data["rotated"], data["trimed"]))


def save_json(data_dict, file_path):
    """
    save a dict as a json file
    :param data_dict: dict data
    :param file_path: json file path to save
    :return:
    """
    import json
    with open(file_path, 'w') as fp:
        json.dump(data_dict, fp)


def save_plist(data_dict, file_path):
    """
    save a dict as a plist file
    :param data_dict: dict data
    :param file_path: plist file path to save
    :return:
    """
    import plistlib

    if hasattr(plistlib, "dump"):
        with open(file_path, 'wb') as fp:
            plistlib.dump(data_dict, fp)
    else:
        plistlib.writePlist(data_dict, file_path)


def save_image(image, file_path):
    """
    save a Image as a file
    :param image: Image
    :param file_path: file path to save
    :return:
    """
    image.save(file_path)


def alpha_bleeding(image, bleeding_pixel=8):
    """
    alpha bleeding
    :param image: Image
    :param bleeding_pixel: pixel to bleed
    :return:
    """
    offsets = ((-1, -1), (0, -1), (1, -1),
               (-1, 0), (1, 0),
               (-1, 1), (0, 1), (1, 1))

    image = image.copy()
    width, height = image.size
    if image.mode != "RGBA":
        image = image.convert("RGBA")
    pa = image.load()

    bleeding = set()
    borders = []

    def _tell_border(x, y):
        if pa[x, y][3] == 0:
            return False

        for offset in offsets:
            ox = x + offset[0]
            oy = y + offset[1]
            if 0 <= ox < width and 0 <= oy < height and pa[ox, oy][3] == 0:
                return True
        return False

    def _bleeding(x, y):
        borders = []
        pixel = pa[x, y]
        for offset in offsets:
            ox = x + offset[0]
            oy = y + offset[1]
            if 0 <= ox < width and 0 <= oy < height and pa[ox, oy][3] == 0 and (ox, oy) not in bleeding:
                pa[ox, oy] = (pixel[0], pixel[1], pixel[2], 1)
                bleeding.add((ox, oy))
                if _tell_border(ox, oy):
                    borders.append((ox, oy))
        return borders

    for x in range(width):
        for y in range(height):
            if _tell_border(x, y):
                borders.append((x, y))

    for i in range(bleeding_pixel):
        pending = []
        for border in borders:
            pending.extend(_bleeding(*border))
        borders = pending

    return image


def alpha_remove(image):
    """
    remove the alpha channel of the image(set as 255)
    :param image: Image
    :return:
    """
    image = image.copy()
    width, height = image.size
    if image.mode != "RGBA":
        image = image.convert("RGBA")
    pa = image.load()
    for x in range(width):
        for y in range(height):
            pixel = pa[x, y]
            pa[x, y] = (pixel[0], pixel[1], pixel[2], 255)
    return image


def clean_pixel_alpha_below(image, v=1):
    """
    clean pixels of the image which alpha channel is below the given value
    :param image: Image
    :param v: given value
    :return:
    """
    image = image.copy()
    width, height = image.size
    if image.mode != "RGBA":
        image = image.convert("RGBA")

    pa = image.load()
    for x in range(width):
        for y in range(height):
            pixel = pa[x, y]
            if pixel[3] < v:
                pa[x, y] = (0, 0, 0, 0)
    return image


def _border(border):
    if isinstance(border, tuple):
        if len(border) == 2:
            left, top = right, bottom = border
        elif len(border) == 4:
            left, top, right, bottom = border
    else:
        left = top = right = bottom = border
    return left, top, right, bottom


def extrude_image(image, border=0):
    if border == 0:
        return image

    from PIL import Image
    left, top, right, bottom = _border(border)
    width = left + image.size[0] + right
    height = top + image.size[1] + bottom
    out = Image.new(image.mode, (width, height))
    out.paste(image, (left, top))
    for x in xrange(width):
        if 0 <= x < left:
            ref_x = left
        elif x >= left + image.size[0]:
            ref_x = left + image.size[0] - 1
        else:
            ref_x = x
        p = out.getpixel((ref_x, top))
        for y in xrange(top):
            out.putpixel((x, y), p)

    for y in xrange(top, top + image.size[1]):
        p = out.getpixel((left, y))
        for x in xrange(left):
            out.putpixel((x, y), p)

    for y in xrange(top, top + image.size[1]):
        p = out.getpixel((left + image.size[0] - 1, y))
        for x in xrange(left + image.size[0], width):
            out.putpixel((x, y), p)

    for x in xrange(width):
        if 0 <= x < left:
            ref_x = left
        elif x >= left + image.size[0]:
            ref_x = left + image.size[0] - 1
        else:
            ref_x = x
        p = out.getpixel((ref_x, top + image.size[1] - 1))
        for y in xrange(top + image.size[1], height):
            out.putpixel((x, y), p)
    return out



# -*- coding: utf-8 -*-
"""----------------------------------------------------------------------------
Author:
    Huang Quanyong (wo1fSea)
    quanyongh@foxmail.com
Date:
    2016/10/19
Description:
    ImageRect.py
----------------------------------------------------------------------------"""

from PIL import Image

from .Rect import Rect
from . import Utils

class ImageRect(Rect):
    """
    image rect data
    """

    def __init__(self, image_path=None):
        super(ImageRect, self).__init__(0, 0, 0, 0)

        self.image = None
        self.image_path = None
        self.source_size = (0, 0)
        self.source_box = (0, 0, 0, 0)

        self._extrude_size = 0
        self._rotated = False
        self._trimmed = False
        if image_path:
            self.load_image(image_path)
            self.image_path = image_path

    @property
    def rotated(self):
        return self._rotated

    @property
    def trimmed(self):
        return self._trimmed

    @property
    def extrude_size(self):
        return self._extrude_size

    @property
    def bbox(self):
        if self._trimmed:
            return self.image.getbbox()
        else:
            return tuple(0, 0, self.width, self.height)

    def load_image(self, image_path):
        image = Image.open(image_path)
        self.image = image.copy()
        image.close()

        self.image_path = image_path

        self.x, self.y = 0, 0
        self.width, self.height = self.image.size

        self.source_size = self.image.size
        self.source_box = (0, 0, self.width, self.height)

        self._rotated = False
        self._trimmed = False

    def rotate(self):
        self._rotated = not self._rotated

        width = self.width
        self.width = self.height
        self.height = width

    def trim(self, v=1):
        if self._trimmed:
            return

        self.image = Utils.clean_pixel_alpha_below(self.image, v)
        bbox = self.image.getbbox()
        if bbox:
            self.image = self.image.crop(bbox)
            self.source_box = bbox
            self.width, self.height = self.image.size

        self._trimmed = True

    def extrude(self, size=0):
        if size <= 0:
            return

        self._extrude_size = size
        self.image = Utils.extrude_image(self.image, self._extrude_size)
        self.width, self.height = self.image.size

    def clone(self):
        tmp = ImageRect()

        tmp.image = self.image
        tmp.image_path = self.image_path

        tmp.x, tmp.y = self.x, self.y
        tmp.width, tmp.height = self.width, self.height

        tmp.source_size = self.source_size
        tmp.source_box = self.source_box

        tmp._rotated = self._rotated
        tmp._trimmed = self._trimmed
        tmp._extrude_size = self._extrude_size
        return tmp


def main():
    img_rect = ImageRect("test.jpg")
    img_rect.rotate()


if __name__ == '__main__':
    main()
